from magma.logging import root_logger


_logger = root_logger()


class WiringLog:
    def __init__(self, tpl, *bits):
        self.tpl = tpl
        self.bits = bits

    def __str__(self):
        bits = [bit.debug_name for bit in self.bits]
        return self.tpl.format(*bits)


class Wire:
    """
    Wire implements wiring.

    Each wire is represented by a bit.
    """
    def __init__(self, bit):
        self._bit = bit
        self._driving = []
        self._driver = None

    def __repr__(self):
        return repr(self._bit)

    def __str__(self):
        return str(self._bit)

    def anon(self):
        return self._bit.anon()

    def unwire(self, other):
        other._driving.remove(self)
        self._driver = None

    def connect(self, other, debug_info):
        """
        Connect two wires, self should be an input and other should be an
        output, or both should be inouts
        """
        if self._driver is not None:
            _logger.warning(
                WiringLog(
                    ("Wiring multiple outputs to same wire, using last "
                     "connection. Input: {}, Old Output: {}, New Output: {}"),
                    self._bit, self._driver._bit, other._bit),
                debug_info=debug_info
            )
        if self._bit.is_output():
            _logger.error(
                WiringLog("Using `{}` (an output) as an input", self._bit),
                debug_info=debug_info
            )
            return
        if other._bit.is_input():
            _logger.error(
                WiringLog("Using `{}` (an input) as an output", other._bit),
                debug_info=debug_info
            )
            return
        if self._bit.is_inout() and not other._bit.is_inout():
            _logger.error(
                WiringLog("Using `{}` (not inout) as an inout", other._bit),
                debug_info=debug_info
            )
            return
        if not self._bit.is_inout() and other._bit.is_inout():
            _logger.error(
                WiringLog("Using `{}` (not inout) as an inout", self._bit),
                debug_info=debug_info
            )
            return

        self._driver = other
        other._driving.append(self)

    def trace(self, skip_self=True):
        """
        If a value is an input or an intermediate (undirected), trace it until
        there is an input or inout (this is the source)

        Upon the first invocation (from a user), we skip the current bit (so
        we don't trace to ourselves)
        """
        if self._driver is not None:
            return self._driver.trace(skip_self=False)
        if not skip_self and (self._bit.is_output() or self._bit.is_inout()):
            return self._bit
        return None

    def value(self):
        """
        Return the bit connected to this bit. Specifically, return the bit this
        bit is driving if it exists. Else if, there is a unique "drivee" bit,
        return that bit. Otherwise, return None.
        """
        if self._driver is not None:
            return self._driver._bit
        if len(self._driving) == 1:
            return self._driving[0]._bit
        return None

    def driven(self):
        return self._driver is not None

    def wired(self):
        return self._driver or self._driving

    def driving(self):
        """
        Return a (possibly empty) list of all bits this bit is driving.
        """
        return [driving._bit for driving in self._driving]

    @property
    def driver(self):
        return self._driver

    @property
    def bit(self):
        return self._bit



import math


def log2_ceil(x: int) -> int:
    return math.ceil(math.log(x, 2))


def log2_floor(x: int) -> int:
    return math.floor(math.log(x, 2))


def is_pow2(x: int) -> bool:
    return log2_ceil(x) == log2_floor(x)



from magma.array import Array
from magma.tuple import Tuple
from magma.protocol_type import magma_type


def wireable(T1, T2):
    """
    Returns true if T1 can be wired to T2
    """
    T1, T2 = magma_type(T1), magma_type(T2)
    if issubclass(T1, Tuple):
        if not issubclass(T2, Tuple):
            return False
        return all(wireable(t1, t2) for t1, t2 in zip(T1, T2))
    if issubclass(T1, Array):
        if not issubclass(T2, Array):
            return False
        return wireable(T1.T, T2.T)
    return issubclass(T1, T2) or issubclass(T1, T2)



import os
from magma.bit import Bit
from magma.bits import Bits
from magma.array import Array
from magma.config import get_compile_dir, set_compile_dir
from magma.digital import Digital
from magma.passes.passes import CircuitPass
from magma.passes.insert_coreir_wires import Wire
from magma.tuple import Tuple
from magma.verilog_utils import value_to_verilog_name, is_nd_array
from magma.t import Direction
from magma.conversions import from_bits, as_bits
from magma.view import PortView, InstView
from magma.wire import wire
from magma.inline_verilog import _get_view_inst_parent


def _gen_bind_port(cls, mon_arg, bind_arg):
    if (isinstance(mon_arg, Tuple) or isinstance(mon_arg, Array) and not
            is_nd_array(type(mon_arg))):
        result = []
        for child1, child2 in zip(mon_arg, bind_arg):
            result += _gen_bind_port(cls, child1, child2)
        return result
    port = value_to_verilog_name(mon_arg, False)
    disable_ndarray = False
    if isinstance(mon_arg, Array) and bind_arg.name.root() is not None:
        # Disable NDArray logic for temporary since CoreIR Wire primitive (used
        # for temporaries) does not support ndarrays yet (root() is not None
        # for Named temporary values)
        disable_ndarray = True
    arg = value_to_verilog_name(bind_arg, disable_ndarray)
    return [(f".{port}({arg})")]


def _wire_temp(bind_arg, temp):
    if bind_arg.is_mixed():
        for x, y in zip(bind_arg, temp):
            _wire_temp(x, y)
        return
    if bind_arg.is_input():
        bind_arg = bind_arg.value()
        if bind_arg is None:
            raise ValueError("Cannot bind undriven input")
    wire(bind_arg, temp)


def _bind(cls, monitor, compile_fn, user_namespace, *args):
    bind_str = monitor.verilogFile
    ports = []
    for mon_arg, cls_arg in zip(monitor.interface.ports.values(),
                                cls.interface.ports.values()):
        if str(mon_arg.name) != str(cls_arg.name):
            error_str = f"""
Bind monitor interface does not match circuit interface
    Monitor Ports: {list(monitor.interface.ports)}
    Circuit Ports: {list(cls.interface.ports)}
"""
            raise TypeError(error_str)
        ports += _gen_bind_port(cls, mon_arg, cls_arg)
    extra_mon_args = list(
        monitor.interface.ports.values()
    )[len(cls.interface):]
    for mon_arg, bind_arg in zip(extra_mon_args, args):
        if isinstance(bind_arg, PortView):
            T = type(bind_arg.port)
            parent = _get_view_inst_parent(bind_arg).parent
            if isinstance(parent, InstView):
                defn = parent.circuit
            else:
                assert isinstance(parent, Circuit)
                defn = type(parent)
            driver = bind_arg.port
        else:
            T = type(bind_arg)
            defn = cls
            driver = bind_arg
        with defn.open():
            if not hasattr(defn, "num_bind_wires"):
                defn.num_bind_wires = 0
            name = f"_magma_bind_wire_{defn.num_bind_wires}"
            defn.num_bind_wires += 1
            temp = T.qualify(Direction.Undirected)(name=name)
            _wire_temp(driver, temp)
            temp.unused()
        if isinstance(bind_arg, PortView):
            temp = PortView[type(temp)](temp, bind_arg.parent.parent)
        bind_arg = temp
        ports += _gen_bind_port(cls, mon_arg, bind_arg)
    ports_str = ",\n    ".join(ports)
    cls_name = cls.name
    monitor_name = monitor.name
    if user_namespace is not None:
        cls_name = user_namespace + "_" + cls_name
        monitor_name = user_namespace + "_" + monitor_name
    bind_str = f"bind {cls_name} {monitor_name} {monitor_name}_inst (\n    {ports_str}\n);"  # noqa
    if not os.path.isdir(".magma"):
        os.mkdir(".magma")
    curr_compile_dir = get_compile_dir()
    set_compile_dir("normal")
    # Circular dependency, need coreir backend to compile, backend imports
    # circuit (for wrap casts logic, we might be able to factor that out).
    compile_fn(f".magma/{monitor.name}", monitor, inline=True,
               user_namespace=user_namespace)
    set_compile_dir(curr_compile_dir)
    with open(f".magma/{monitor.name}.v", "r") as f:
        content = "\n".join((f.read(), bind_str))
    cls.compiled_bind_modules[monitor.name] = content


class BindPass(CircuitPass):
    def __init__(self, main, compile_fn, user_namespace):
        super().__init__(main)
        self._compile_fn = compile_fn
        self.user_namespace = user_namespace

    def __call__(self, cls):
        for monitor, args in cls.bind_modules.items():
            _bind(cls, monitor, self._compile_fn, self.user_namespace, *args)



"""
Definition of magma's Bit type
* Subtype of the Digital type
* Implementation of hwtypes.AbstractBit
"""
import keyword
import operator
import typing as tp
import functools
from functools import lru_cache
import hwtypes as ht
from hwtypes.bit_vector_abc import AbstractBit, TypeFamily
from .t import Direction, In, Out
from .digital import Digital, DigitalMeta
from .digital import VCC, GND  # TODO(rsetaluri): only here for b.c.

from magma.compatibility import IntegerTypes
from magma.circuit import Circuit, coreir_port_mapping
from magma.debug import debug_wire
from magma.family import get_family
from magma.interface import IO
from magma.language_utils import primitive_to_python
from magma.operator_utils import output_only


def bit_cast(fn: tp.Callable[['Bit', 'Bit'], 'Bit']) -> \
        tp.Callable[['Bit', tp.Union['Bit', bool]], 'Bit']:
    @functools.wraps(fn)
    def wrapped(self: 'Bit', other: tp.Union['Bit', bool]) -> 'Bit':
        if isinstance(other, Bit):
            return fn(self, other)
        try:
            other = Bit(other)
        except TypeError:
            return NotImplemented
        return fn(self, other)
    return wrapped


_IMPLICITLY_COERCED_ITE_TYPES = (int, ht.BitVector, ht.Bit)


class Bit(Digital, AbstractBit, metaclass=DigitalMeta):
    __hash__ = Digital.__hash__

    @staticmethod
    def get_family() -> TypeFamily:
        return get_family()

    @classmethod
    @lru_cache(maxsize=None)
    def declare_unary_op(cls, op):
        assert op == "not", f"simulate not implemented for {op}"

        class _MagmaBitOp(Circuit):
            name = f"magma_Bit_{op}"
            coreir_name = op
            coreir_lib = "corebit"
            renamed_ports = coreir_port_mapping
            primitive = True
            stateful = False

            io = IO(I=In(Bit), O=Out(Bit))

            def simulate(self, value_store, state_store):
                I = ht.Bit(value_store.get_value(self.I))
                O = int(~I)
                value_store.set_value(self.O, O)

        return _MagmaBitOp

    @classmethod
    @lru_cache(maxsize=None)
    def declare_binary_op(cls, op):
        python_op_name = primitive_to_python(op)
        python_op = getattr(operator, python_op_name)

        class _MagmaBitOp(Circuit):
            name = f"magma_Bit_{op}"
            coreir_name = op
            coreir_lib = "corebit"
            renamed_ports = coreir_port_mapping
            primitive = True
            stateful = False

            io = IO(I0=In(Bit), I1=In(Bit), O=Out(Bit))

            def simulate(self, value_store, state_store):
                I0 = ht.Bit(value_store.get_value(self.I0))
                I1 = ht.Bit(value_store.get_value(self.I1))
                O = int(python_op(I0, I1))
                value_store.set_value(self.O, O)

        return _MagmaBitOp

    def __init__(self, value=None, name=None):
        super().__init__(name=name)
        if value is None:
            self._value = None
        elif isinstance(value, Bit):
            self._value = value._value
        elif isinstance(value, bool):
            self._value = value
        elif isinstance(value, int):
            if value not in {0, 1}:
                raise ValueError('Bit must have value 0 or 1 not {}'.format(value))
            self._value = bool(value)
        elif hasattr(value, '__bool__'):
            self._value = bool(value)
        else:
            raise TypeError("Can't coerce {} to Bit".format(type(value)))

    @property
    def direction(self):
        return type(self).direction

    def __invert__(self):
        # CoreIR uses not instead of invert name
        return self.declare_unary_op("not")()(self)

    @bit_cast
    @output_only("Cannot use == on an input")
    def __eq__(self, other):
        # CoreIR doesn't define an eq primitive for bits
        return ~(self ^ other)

    @bit_cast
    @output_only("Cannot use != on an input")
    def __ne__(self, other):
        # CoreIR doesn't define an ne primitive for bits
        return self ^ other

    @bit_cast
    def __and__(self, other):
        return self.declare_binary_op("and")()(self, other)

    @bit_cast
    def __or__(self, other):
        return self.declare_binary_op("or")()(self, other)

    @bit_cast
    def __xor__(self, other):
        return self.declare_binary_op("xor")()(self, other)

    def ite(self, t_branch, f_branch):
        if isinstance(t_branch, list) and isinstance(f_branch, list):
            if not len(t_branch) == len(f_branch):
                raise TypeError("Bit.ite of two lists expects the same length")
            return [self.ite(x, y) for x, y in zip(t_branch, f_branch)]
        t_type = type(t_branch)
        f_type = type(f_branch)

        # Implicit int/bv conversion
        if (issubclass(t_type, _IMPLICITLY_COERCED_ITE_TYPES) and
                not issubclass(f_type, _IMPLICITLY_COERCED_ITE_TYPES)):
            t_branch = f_type(t_branch)
            t_type = f_type
        if (not issubclass(t_type, _IMPLICITLY_COERCED_ITE_TYPES) and
                issubclass(f_type, _IMPLICITLY_COERCED_ITE_TYPES)):
            f_branch = t_type(f_branch)
            f_type = t_type

        # allows undirected types to match (e.g. for temporary values)
        if (t_type is not f_type and
                t_type.qualify(Direction.Undirected) is not f_type and
                f_type.qualify(Direction.Undirected) is not t_type):
            while True:
                try:
                    if t_type.is_wireable(f_type):
                        break
                except AttributeError:
                    pass
                try:
                    if f_type.is_wireable(t_type):
                        break
                except AttributeError:
                    pass
                raise TypeError("ite expects same type for both branches: "
                                f"{t_type} != {f_type}")

        if self.const():
            if self is type(self).VCC:
                return t_branch
            assert self is type(self).GND
            return f_branch
        if issubclass(t_type, tuple):
            return tuple(self.ite(t, f) for t, f in zip(t_branch, f_branch))
        # Note: coreir flips t/f cases
        # self._Mux monkey patched in magma/primitives/mux.py to avoid circular
        # dependency
        return self._Mux(2, t_type)()(f_branch, t_branch, self)

    def __bool__(self) -> bool:
        raise NotImplementedError("Converting magma bit to bool not supported")

    def __int__(self) -> int:
        raise NotImplementedError("Converting magma bit to int not supported")

    def unused(self):
        if self.is_input() or self.is_inout():
            raise TypeError("unused cannot be used with input/inout")
        if not getattr(self, "_magma_unused_", False):
            DefineUnused()().I.wire(self)
            # "Cache" unused calls so only one is produced
            self._magma_unused_ = True

    def undriven(self):
        if self.is_output() or self.is_inout():
            raise TypeError("undriven cannot be used with output/inout")
        self.wire(DefineUndriven()().O)

    @debug_wire
    def wire(self, o, debug_info):
        # Cast to Bit here so we don't get a Digital instead
        if isinstance(o, (IntegerTypes, bool, ht.Bit)):
            o = Bit(o)
        return super().wire(o, debug_info)


def make_Define(_name, port, direction):
    @lru_cache(maxsize=None)
    def DefineCorebit():
        class _Primitive(Circuit):
            renamed_ports = coreir_port_mapping
            name = f"corebit_{_name}"
            coreir_name = _name
            coreir_lib = "corebit"

            def simulate(self, value_store, state_store):
                pass

            # Type must be a bit because coreir uses Bit for the primitive.
            io = IO(**{port: direction(Bit)})
        return _Primitive
    return DefineCorebit


DefineUndriven = make_Define("undriven", "O", Out)
DefineUnused = make_Define("term", "I", In)

# Hack to avoid circular dependency
Digital.unused = Bit.unused
Digital.undriven = Bit.undriven

BitIn = Bit[Direction.In]
BitOut = Bit[Direction.Out]
BitInOut = Bit[Direction.InOut]



# https://wiki.python.org/moin/BitManipulation
# https://code.google.com/p/python-bitstring/
from types import FunctionType
from collections.abc import Sequence
import inspect
from .compatibility import StringTypes

__all__  = ['seq2int', 'int2seq', 'ints2seq', 'fun2seq', 'lutinit']
__all__ += ['int2uint', 'clz', 'pow2', 'log2', 'clog2', 'rol', 'ror']

#
# seq to int
#
def seq2int(l):
    n = len(l)

    i = 0
    for j in range(n):
        if l[j]:
            i |= 1 << j
    return i

#
# int to seq
#
def int2seq(i, n=0):
    if isinstance(i, StringTypes):
        i = ord(i)

    # find minimum number of bits needed for i
    if n == 0:
        j = i
        while j:
            n += 1
            j >>= 1

    return [1 if i & (1 << j) else 0 for j in range(n)]

#
# ints to seq
#
def ints2seq(l, n):
    return [int2seq(i,n) for i in l]

#
# fun to seq
#
def fun2seq(f, n=None):
    if not n:
        logn = len(inspect.getargspec(f).args)
        n = 1 << logn
    else:
        logn = log2(n)
    
    l = []
    for i in range(n):
         arg = int2seq(i, logn)
         l.append(1 if f(*arg) else 0)
    return l

def lutinit(init, n=None):
    if isinstance(init, FunctionType):
        init = fun2seq(init, n)

    if isinstance(init, Sequence):
        nlut = len(init)
        if n != nlut:
            assert n % nlut == 0
            init = (n//nlut) * init
        init = seq2int(init)

    assert n is not None

    return init, n


def int2uint(x, nbits):
    return x & ((1 << nbits)-1)

# base 32
def clz(x):
    if x == 0:
        return 32

    n = 0
    if x <= 0x0000FFFF:
        n = n + 16
        x = x << 16
    if x <= 0x00FFFFFF:
        n = n + 8
        x = x << 8
    if x <= 0x0FFFFFFF:
        n = n + 4
        x = x << 4
    if x <= 0x3FFFFFFF:
        n = n + 2
        x = x << 2
    if x <= 0x7FFFFFFF:
        n = n + 1
    return n


def log2(x):
    return 31 - clz(x)

def clog2(x):
    if x == 0: return 0
    y = log2(x)
    if x > (1 << y):
        y = y + 1
    return y

def pow2(n):
    return 1 << n

def rol(l, m):
    n = len(l)
    return l[m:n] + l[0:m]

def ror(l, m):
    n = len(l)
    return l[n-m:n] + l[0:n-m]


if __name__ == "__main__":
    def f(a,b):
        return a or b
    print(fun2seq(f))



from typing import Union

import coreir
import hwtypes as ht

from magma.array import Array
from magma.bit import Bit
from magma.clock import Enable
from magma.common import ParamDict
from magma.bits import Bits, UInt, SInt
from magma.circuit import coreir_port_mapping
from magma.conversions import as_bits, from_bits, bit
from magma.interface import IO
from magma.generator import Generator2
from magma.t import Type, Kind, In, Out, Direction
from magma.tuple import Tuple
from magma.clock import (AbstractReset,
                         AsyncReset, AsyncResetN, Clock, get_reset_args)
from magma.clock_io import ClockIO
from magma.primitives.mux import Mux
from magma.wire import wire
from magma.wireable import wireable


class _CoreIRRegister(Generator2):
    """
    Internally used generator for CoreIR register primitive
    """
    def __init__(self, width, init=0, has_async_reset=False,
                 has_async_resetn=False):
        self.name = "reg_P"
        self.coreir_configargs = {"init": coreir.type.BitVector[width](init)}
        T = Bits[width]
        self.io = IO(I=In(T), CLK=In(Clock), O=Out(T))

        if has_async_resetn and has_async_reset:
            raise ValueError("Cannot have posedge and negedge asynchronous "
                             "reset")

        if has_async_reset:
            self.io += IO(arst=In(AsyncReset))
            self.name += "R"
            self.coreir_configargs["arst_posedge"] = True

        if has_async_resetn:
            self.io += IO(arst=In(AsyncResetN))
            self.name += "R"
            self.coreir_configargs["arst_posedge"] = False

        self.stateful = True
        self.default_kwargs = {"init": coreir.type.BitVector[width](init)}
        self.coreir_genargs = {"width": width}
        self.renamed_ports = coreir_port_mapping

        self.coreir_name = "reg"
        if (has_async_reset or has_async_resetn):
            self.coreir_name += "_arst"

        self.verilog_name = "coreir_" + self.coreir_name
        self.coreir_lib = "coreir"

        def _simulate(self, value_store, state_store):
            cur_clock = value_store.get_value(self.CLK)

            if not state_store:
                state_store['prev_clock'] = cur_clock
                state_store['cur_val'] = ht.BitVector[width](init)

            if has_async_reset or has_async_resetn:
                cur_reset = value_store.get_value(self.arst)

            prev_clock = state_store['prev_clock']
            clock_edge = cur_clock and not prev_clock

            new_val = state_store['cur_val'].as_bool_list()

            if clock_edge:
                new_val = value_store.get_value(self.I)

            if has_async_reset and cur_reset or \
                    has_async_resetn and not cur_reset:
                new_val = ht.BitVector[width](init)

            state_store['prev_clock'] = cur_clock
            state_store['cur_val'] = ht.BitVector[width](new_val)
            value_store.set_value(self.O, new_val)
        self.simulate = _simulate


def _zero_init(T, init):
    if issubclass(T, Bits):
        return T(init)
    elif issubclass(T, Array):
        return T([_zero_init(T.T, init) for _ in range(len(T))])
    elif issubclass(T, Tuple):
        return T(*(_zero_init(t, init) for t in T.types()))
    return T(init)


def _get_T_from_init(init):
    if isinstance(init, (bool, ht.Bit)):
        return Bit
    if isinstance(init, int):
        return Bits[max(init.bit_length(), 1)]
    if isinstance(init, ht.UIntVector):
        return UInt[len(init)]
    if isinstance(init, ht.SIntVector):
        return SInt[len(init)]
    if isinstance(init, ht.BitVector):
        return Bits[len(init)]
    if isinstance(init, Type):
        return type(init)
    raise ValueError("Could not infer register type from {init}")


def _check_init_T(init, T):
    init_T = _get_T_from_init(init)
    if isinstance(init, int) and issubclass(T, Bits):
        # Allow int to be extended to width of T
        if len(init_T) > len(T):
            # Don't implicitly truncate
            return False
        return True
    if isinstance(init, int) and issubclass(T, Bit):
        # Allow int for bit
        if len(init_T) > 1:
            return False
        return True
    return wireable(init_T, T)


class Register(Generator2):
    def __init__(self, T: Kind = None,
                 init: Union[Type, int, ht.BitVector] = None,
                 reset_type: AbstractReset = None, has_enable: bool = False,
                 reset_priority: bool = True, name_map=ParamDict(I="I",
                                                                 O="O",
                                                                 CE="CE")):
        """
        T: The type of the value that is stored inside the register (e.g.
           Bits[5])

        init: (optional) A const value (i.e. init.const() == True) of type T or
              an int to be used as the initial value of the register.
              If no value is provided, the register will be initialized with 0

        has_enable: (optional) whether the register has an enable signal

        reset_type: (optional) The type of the reset port (also specifies the
                    semantic behavior of the reset signal)

        reset_priority: (optional) boolean flag choosing whether synchronous
                        reset (RESET or RESETN) has priority over enable

        name_map: (optional) ParamDict mapping default port names to new names
        """
        if T is None:
            if init is None:
                raise ValueError("User must provide type T or init value (from"
                                 " which T will be inferred)")
            T = _get_T_from_init(init)
        else:
            if not isinstance(T, Kind):
                raise TypeError(
                    f"Expected instance of Kind for argument T, not {type(T)}")
            if init is not None and not _check_init_T(init, T):
                raise ValueError(
                    f"Type {_get_T_from_init(init)} of init ({init}) does not "
                    f"match T ({T})"
                )

        (
            has_async_reset, has_async_resetn, has_reset, has_resetn
        ) = get_reset_args(reset_type)

        I_name = name_map.get("I", "I")
        O_name = name_map.get("O", "O")
        CE_name = name_map.get("CE", "CE")
        self.io = IO(**{I_name: In(T), O_name: Out(T)})
        if has_enable:
            self.io += IO(**{CE_name: In(Enable)})
        self.io += ClockIO(has_async_reset=has_async_reset,
                           has_async_resetn=has_async_resetn,
                           has_reset=has_reset,
                           has_resetn=has_resetn)
        if init is None:
            init = 0

        if isinstance(init, int):
            init = _zero_init(T, init)

        coreir_init = init
        if isinstance(coreir_init, Type):
            coreir_init = as_bits(init)
        coreir_init = int(coreir_init)

        reg = _CoreIRRegister(T.flat_length(), init=coreir_init,
                              has_async_reset=has_async_reset,
                              has_async_resetn=has_async_resetn)()
        O = from_bits(T, reg.O)
        wire(O, getattr(self.io, O_name))

        I = getattr(self.io, I_name)
        if has_reset:
            reset_select = self.io.RESET
        elif has_resetn:
            reset_select = ~bit(self.io.RESETN)
        if has_enable:
            enable = getattr(self.io, CE_name)
        if (has_reset or has_resetn) and has_enable:
            if reset_priority:
                I = Mux(2, T)(name="enable_mux")(O, I, enable)
                I = Mux(2, T)()(I, init, reset_select)
            else:
                I = Mux(2, T)()(I, init, reset_select)
                I = Mux(2, T)(name="enable_mux")(O, I, enable)
        elif has_enable:
            I = Mux(2, T)(name="enable_mux")(O, I, enable)
        elif (has_reset or has_resetn):
            I = Mux(2, T)()(I, init, reset_select)
        reg.I @= as_bits(I)


def register(value, **kwargs):
    T = type(value).qualify(Direction.Undirected)
    inst_kwargs = {}
    try:
        name = kwargs.pop("name")
        inst_kwargs["name"] = name
    except KeyError:
        pass
    ckt = Register(T, **kwargs)
    return ckt(**inst_kwargs)(value)



import weakref
from functools import reduce
from abc import ABCMeta
from hwtypes import BitVector
from .common import deprecated
from .ref import AnonRef, ArrayRef
from .t import Type, Kind, Direction, In, Out
from .compatibility import IntegerTypes
from .digital import Digital
from .bit import Bit
from .bitutils import int2seq, seq2int
from .debug import debug_wire, get_callee_frame_info
from .logging import root_logger
from .protocol_type import magma_type, magma_value

from magma.operator_utils import output_only
from magma.wire_container import WiringLog
from magma.protocol_type import MagmaProtocol


_logger = root_logger()


class ArrayMeta(ABCMeta, Kind):
    # BitVectorType, size :  BitVectorType[size]
    _class_cache = weakref.WeakValueDictionary()

    def __new__(cls, name, bases, namespace, info=(None, None, None), **kwargs):
        # TODO: A lot of this code is shared with AbstractBitVectorMeta, we
        # should refactor to reuse
        if '_info_' in namespace:
            raise TypeError(
                'class attribute _info_ is reversed by the type machinery')

        N, T = info[1:3]
        for base in bases:
            if getattr(base, 'is_concrete', False):
                if (N, T) == (None, None):
                    (N, T) = (base.N, base.T)
                elif N != base.N:
                    raise TypeError(
                        "Can't inherit from multiple arrays with different N")
                elif not issubclass(T, base.T):
                    raise TypeError(
                        "Can't inherit from multiple arrays with different T")

        namespace['_info_'] = info[0], N, T
        type_ = super().__new__(cls, name, bases, namespace, **kwargs)
        if (N ,T) == (None, None):
            #class is abstract so t.abstract -> t
            type_._info_ = type_, N, T
        elif info[0] is None:
            #class inherited from concrete type so there is no abstract t
            type_._info_ = None, N, T

        return type_

    def __getitem__(cls, index: tuple) -> 'ArrayMeta':
        mcs = type(cls)

        # If cls.T is a direction, and the new T (index[1])
        if isinstance(cls.T, Direction):
            # If we're replacing a direction (e.g. `In(Out(Array)))`, just use
            # the default direction logic
            if not isinstance(index[1], Direction):
                # Otherwise, we expect that we're qualifying a Type with the
                # direction (e.g. In(Array)[5, Bit])
                if not issubclass(index[1], Type):
                    raise TypeError("Expected Type as second index to Array")
                if not index[1].is_oriented(cls.T):
                    _logger.warning(
                        f"Parametrizing qualifed Array {cls} with inner type "
                        f" {index[1]} which doesn't match, will use array "
                        "qualifier"
                    )
                index = index[0], index[1].qualify(cls.T)
        elif cls.T is None:
            # Else, it index[1] should be  Type (e.g. In(Bit)) or a Direction
            # (used internally for In(Array))
            valid_second_index = (isinstance(index[1], Direction) or
                                  issubclass(magma_type(index[1]), Type))
            if not valid_second_index:
                raise TypeError(
                    "Expected Type or Direction as second index to Array"
                    f" got: {index[1], type(index[1])}"
                )

        try:
            return mcs._class_cache[cls, index]
        except KeyError:
            pass

        if not (isinstance(index, tuple) and len(index) == 2):
            raise TypeError('Parameters to array must be a tuple of length 2')

        # index[0] (N) can be None (used internally for In(Array))
        if index[0] is not None:
            if isinstance(index[0], tuple):
                if len(index[0]) == 0 :
                    raise ValueError("Cannot create array with length 0 tuple "
                                     "for N")
                if len(index[0]) > 1:
                    T = index[1]
                    # ND Array
                    for N in index[0]:
                        T = Array[N, T]
                    return T
                # len(index[0]) == 1, Treat as normal Array
                index = index[0]

            if (not isinstance(index[0], int) or index[0] <= 0):
                raise TypeError(
                    'Length of array must be an int greater than 0, got:'
                    f' {index[0]}'
                )


        if cls.is_concrete:
            if index[0] == cls.N and index[1] is cls.T:
                return cls
            else:
                return cls.abstract_t[index]

        bases = []
        bases.extend(b[index] for b in cls.__bases__ if isinstance(b, mcs))
        # only add base classes if we're have a child type
        # (skipped in the case of In(Array))
        if not isinstance(index[1], Direction):
            bases.extend(cls[index[0], b] for b in index[1].__bases__ if
                         isinstance(b, type(magma_type(index[1]))))
        if not any(issubclass(b, cls) for b in bases):
            bases.insert(0, cls)
        bases = tuple(bases)
        orig_name = cls.__name__
        if isinstance(index[1], Direction):
            class_name = f'{index[1].name}({cls.__name__})'
        else:
            class_name = '{}[{}]'.format(cls.__name__, index)
        type_ = mcs(class_name, bases, {"orig_name": orig_name}, info=(cls, ) + index)
        type_.__module__ = cls.__module__
        mcs._class_cache[cls, index] = type_
        return type_

    @property
    def abstract_t(cls) -> 'ArrayMeta':
        t = cls._info_[0]
        if t is not None:
            return t
        else:
            raise AttributeError('type {} has no abstract_t'.format(cls))

    @property
    def undirected_t(cls) -> 'ArrayMeta':
        T = cls.T
        if cls.is_concrete:
            return cls[cls.N, cls.T.qualify(Direction.Undirected)]
        else:
            raise AttributeError('type {} has no undirected_t'.format(cls))

    @property
    def N(cls) -> int:
        return cls._info_[1]

    @property
    def T(cls):
        return cls._info_[2]

    @property
    def is_concrete(cls) -> bool:
        return (cls.N, cls.T) != (None, None)

    def __len__(cls):
        return cls.N

    def __str__(cls):
        # handle In(Array)
        if isinstance(cls.T, Direction):
            assert cls.N is None
            return f"{cls.T.name}(Array)"
        return f"Array[{cls.N}, {cls.T}]"

    def __repr__(cls):
        return f"Array[{cls.N}, {cls.T}]"

    def qualify(cls, direction):
        # Handle qualified, unsized/child e.g. In(Array) and In(Out(Array))
        if cls.T is None or isinstance(cls.T, Direction):
            return cls[None, direction]
        return cls[cls.N, cls.T.qualify(direction)]

    def flip(cls):
        return cls[cls.N, cls.T.flip()]

    def __eq__(cls, rhs):
        if not isinstance(rhs, ArrayMeta):
            return NotImplemented
        return (cls.N == rhs.N) and (cls.T == rhs.T)

    def is_wireable(cls, rhs):
        rhs = magma_type(rhs)
        if not isinstance(rhs, ArrayMeta) or cls.N != rhs.N:
            return False
        return cls.T.is_wireable(rhs.T)

    def is_bindable(cls, rhs):
        rhs = magma_type(rhs)
        if not isinstance(rhs, ArrayMeta) or cls.N != rhs.N:
            return False
        return cls.T.is_bindable(rhs.T)

    __hash__ = type.__hash__


def _is_valid_slice(N, key):
    start, stop = key.start, key.stop
    return (((start is None or (start < N and start >= -N))) and
            (stop is None or (stop <= N and stop > -N)))


class Array(Type, metaclass=ArrayMeta):
    def __init__(self, *args, **kwargs):
        Type.__init__(self, **kwargs)
        self.ts = []
        if args:
            if len(args) == 1 and isinstance(args[0], (list, Array, int,
                                                       BitVector)):
                if isinstance(args[0], list):
                    if len(args[0]) != self.N:
                        raise ValueError("Array list constructor can only be used "
                                         "with list equal to array length")
                    self.ts = []
                    for elem in args[0]:
                        if isinstance(elem, int):
                            self.ts.append(self.T(elem))
                        else:
                            self.ts.append(elem)
                elif isinstance(args[0], Array):
                    if len(args[0]) != len(self):
                        raise TypeError(f"Will not do implicit conversion of arrays")
                    self.ts = args[0].ts[:]
                elif isinstance(args[0], (BitVector, int)):
                    if not issubclass(self.T, Bit):
                        raise TypeError(f"Can only instantiate Array[N, Bit] "
                                        f"with int/bv, not Array[N, {self.T}]")
                    if (isinstance(args[0], BitVector) and
                            len(args[0]) != self.N):
                        raise TypeError(
                            f"Cannot create a value of type {type(self)} with"
                            f" a BitVector of length {len(args[0])} (the sizes"
                            " must match)"
                        )

                    n = args[0]
                    bits = (int2seq(n, self.N)
                            if isinstance(n, int)
                            else n.bits())
                    self.ts = []
                    for bit in bits:
                        self.ts.append(self.T(bit))
                elif self.N == 1:
                    t = args[0]
                    if isinstance(t, IntegerTypes):
                        t = self.T(t)
                    assert type(t) == self.T or type(t) == self.T.flip() or \
                        issubclass(type(type(t)), type(self.T)) or \
                        issubclass(type(self.T), type(type(t))), (type(t), self.T)
                    self.ts = [t]
            elif len(args) == self.N:
                self.ts = []
                for t in args:
                    if isinstance(t, IntegerTypes):
                        t = self.T(t)
                    assert type(t) == self.T or type(t) == self.T.flip() or \
                        issubclass(type(type(t)), type(self.T)) or \
                        issubclass(type(self.T), type(type(t))), (type(t), self.T)
                    self.ts.append(t)
            else:
                raise TypeError(f"Constructing array with {args} not supported")
        else:
            for i in range(self.N):
                T = self.T
                ref = ArrayRef(self, i)
                if issubclass(T, MagmaProtocol):
                    t = T._from_magma_value_(T._to_magma_()(name=ref))
                else:
                    t = T(name=ref)
                self.ts.append(t)

    @classmethod
    def is_oriented(cls, direction):
        if cls.T is None:
            return False
        return cls.T.is_oriented(direction)

    @classmethod
    def is_clock(cls):
        return False

    @classmethod
    @deprecated
    def isoriented(cls, direction):
        return cls.is_oriented(direction)

    @output_only("Cannot use == on an input")
    def __eq__(self, rhs):
        if not isinstance(rhs, ArrayType):
            return False
        return self.ts == rhs.ts

    @output_only("Cannot use != on an input")
    def __ne__(self, rhs):
        return ~(self == rhs)

    __hash__ = Type.__hash__

    def __repr__(self):
        if self.name.anon():
            t_strs = ', '.join(repr(t) for t in self.ts)
            return f'array([{t_strs}])'
        return super().__repr__()

    @property
    def T(self):
        return type(self).T

    @property
    def N(self):
        return type(self).N

    def __len__(self):
        return self.N

    @classmethod
    def flat_length(cls):
        return cls.N * cls.T.flat_length()

    def _is_whole_slice(self, key):
        # check if it's any of `x[:], x[0:], x[:len(x)], x[0:len(x)]`
        return (isinstance(key[-1], slice) and
                (key[-1] == slice(None) or
                 key[-1] == slice(0, None) or
                 key[-1] == slice(None, len(self)) or
                 key[-1] == slice(0, len(self))))

    def __getitem__(self, key):
        if isinstance(key, tuple):
            # ND Array key
            if len(key) == 1:
                return self[key[0]]

            if not isinstance(key[-1], slice):
                return self[key[-1]][key[:-1]]
            if not self._is_whole_slice(key):
                # If it's not a slice of the whole array, first slice the
                # current array (self), then replace with a slice of the whole
                # array (this is how we determine that we're ready to traverse
                # into the children)
                this_key = key[-1]
                result = self[this_key][key[:-1] + (slice(None), )]
                return result
            # Last index is selecting the whole array, recurse into the
            # children and slice off the inner indices
            inner_ts = [t[key[:-1]] for t in self.ts]
            # Get the type from the children and return the final value
            return type(self)[len(self), type(inner_ts[0])](inner_ts)
        if isinstance(key, Type):
            # indexed using a dynamic magma value, generate mux circuit
            return self.dynamic_mux_select(key)
        if isinstance(key, slice):
            if not _is_valid_slice(self.N, key):
                raise IndexError(f"array index out of range "
                                 f"(type={type(self)}, key={key})")
            _slice = [self[i] for i in range(*key.indices(len(self)))]
            return type(self)[len(_slice), self.T](_slice)
        else:
            if isinstance(key, BitVector):
                key = key.as_uint()

            if not (-self.N <= key and key < self.N):
                raise IndexError(f"{key}, {self.N}")

            return self.ts[key]

    def __setitem__(self, key, val):
        error = False
        old = self[key]
        if isinstance(old, Array):
            if len(old) != len(val):
                error = True
            elif issubclass(old.T, Array):
                # If array of array, check that we can do elementwise setitem
                # (will return true if there's an error)
                # We can't just do an `is` check on the children since those
                # might be slices that return new anon values (so x[1:2] is not
                # x[1:2], but their recursive leaf contents should be the same)
                error = any(old.__setitem__(i, val[i])
                            for i in range(len(old)))
            elif any(old[i] is not val[i] for i in range(len(old))):
                error = True
        elif old is not val:
            error = True

        if error:
            _logger.error(
                WiringLog(f"May not mutate array, trying to replace "
                          f"{{}}[{key}] ({{}}) with {{}}", self, old, val)
            )
        return error

    def __add__(self, other):
        other_len = other.N
        total = self.N + other_len
        res_bits = []
        for i in range(total):
            res_bits.append(self[i] if i < self.N else other[i - self.N])
        return type(self)[len(res_bits), self.T](res_bits)

    def __call__(self, o):
        return self.wire(o, get_callee_frame_info())

    @classmethod
    def is_oriented(cls, direction):
        return cls.T.is_oriented(direction)

    def as_list(self):
        return [self[i] for i in range(len(self))]

    @debug_wire
    def wire(i, o, debug_info):
        if not isinstance(o, ArrayType):
            if isinstance(o, IntegerTypes):
                _logger.error(
                    WiringLog(f"Cannot wire {o} (type={type(o)}) to {{}} "
                              f"(type={type(i)}) because conversions from "
                              f"IntegerTypes are only defined for Bits, not "
                              f"general Arrays", i),
                    debug_info=debug_info
                )
            else:
                o_str = getattr(o, "debug_name", str(o))
                _logger.error(
                    WiringLog(f"Cannot wire {{}} (type={type(o)}) to {{}} "
                              f"(type={type(i)}) because {{}} is not an Array",
                              o, i, o),
                    debug_info=debug_info
                )
            return

        if i.N != o.N:
            _logger.error(
                WiringLog(f"Cannot wire {{}} (type={type(o)}) to {{}} "
                          f"(type={type(i)}) because the arrays do not have "
                          f"the same length", o, i),
                debug_info=debug_info
            )
            return

        for k in range(len(i)):
            i[k].wire(o[k], debug_info)

    def unwire(i, o):
        for k in range(len(i)):
            i[k].unwire(o[k])

    def driven(self):
        for t in self.ts:
            if not t.driven():
                return False
        return True

    def driving(self):
        return [t.driving() for t in self]

    def wired(self):
        for t in self.ts:
            if not t.wired():
                return False
        return True

    # test whether the values refer a whole array
    @staticmethod
    def _iswhole(ts):

        n = len(ts)

        for i in range(n):
            if ts[i].anon():
                return False

        for i in range(n):
            # elements must be an array reference
            if not isinstance(ts[i].name, ArrayRef):
                return False

        for i in range(1, n):
            # elements must refer to the same array
            if ts[i].name.array is not ts[i-1].name.array:
                return False

        if n > 0 and n != ts[0].name.array.N:
            # must use all of the elements of the base array
            return False

        for i in range(n):
            # elements should be numbered consecutively
            if ts[i].name.index != i:
                return False

        return True

    def iswhole(self):
        return Array._iswhole(self.ts)

    def trace(self):
        ts = [t.trace() for t in self.ts]

        for t in ts:
            if t is None:
                return None

        if Array._iswhole(ts):
            return ts[0].name.array

        return type(self).flip()(ts)

    def value(self):
        ts = [t.value() for t in self.ts]

        for t in ts:
            if t is None:
                return None

        if Array._iswhole(ts):
            return ts[0].name.array

        return type(self).flip()(ts)

    def const(self):
        for t in self.ts:
            if not t.const():
                return False

        return True

    @classmethod
    def unflatten(cls, value):
        size_T = cls.T.flat_length()
        if len(value) != size_T * cls.N:
            raise TypeError("Width mismatch")
        ts = [cls.T.unflatten(value[i:i + size_T])
              for i in range(0, size_T * cls.N, size_T)]
        return cls(ts)

    def flatten(self):
        return sum([t.flatten() for t in self.ts], [])

    def concat(self, other) -> 'AbstractBitVector':
        return type(self)[len(self) + len(other), self.T](self.ts + other.ts)

    def undriven(self):
        for elem in self:
            elem.undriven()

    def unused(self):
        for elem in self:
            elem.unused()

    @classmethod
    def is_mixed(cls):
        return cls.T.is_mixed()


ArrayType = Array



import abc
import enum
from magma.common import deprecated
from magma.compatibility import IntegerTypes, StringTypes
from magma.ref import AnonRef, NamedRef, DefnRef, InstRef
from magma.protocol_type import magma_value


class Direction(enum.Enum):
    In = 0
    Out = 1
    InOut = 2
    Undirected = 3


class Type(object):
    def __init__(self, name=None):
        if name is None:
            name = AnonRef()
        elif isinstance(name, str):
            name = NamedRef(name=name, value=self)
        self.name = name

    __hash__ = object.__hash__

    def __repr__(self):
        if self.name.anon():
            return f"{type(self)}()"
        has_name = (isinstance(self.name, NamedRef) and
                    not isinstance(self.name, (InstRef, DefnRef)))
        if has_name:
            return f"{type(self)}(name=\"{repr(self.name)}\")"
        return repr(self.name)

    def __str__(self):
        if self.name.anon():
            # Anon names aren't very useful, so just return a repr instead so
            # it's easier to find the value
            return repr(self)
        return str(self.name)

    # An instance has an anon name.
    def anon(self):
        return self.name.anon()

    # Abstract method to be implemented by subclasses.
    @classmethod
    def is_oriented(cls, direction):
        raise NotImplementedError()

    @classmethod
    def is_clock(cls):
        raise NotImplementedError()

    @classmethod
    def is_input(cls):
        return cls.is_oriented(Direction.In)

    @classmethod
    def is_output(cls):
        return cls.is_oriented(Direction.Out)

    @classmethod
    def is_inout(cls):
        return cls.is_oriented(Direction.InOut)

    @classmethod
    @deprecated
    def isoriented(cls, direction):
        return cls.is_oriented(direction)

    @classmethod
    @deprecated
    def isinput(cls):
        return cls.is_input()

    @classmethod
    @deprecated
    def isoutput(cls):
        return cls.is_output()

    @classmethod
    @deprecated
    def isinout(cls):
        return cls.is_inout()

    @property
    def debug_name(self):
        defn_str = ""
        inst_str = ""
        if isinstance(self.name, DefnRef):
            defn_str = str(self.name.defn.name) + "."
        elif isinstance(self.name, InstRef):
            inst_str = str(self.name.inst.name) + "."
            defn_str = str(self.name.inst.defn.name) + "."
        return f"{defn_str}{inst_str}{str(self)}"

    def __le__(self, other):
        if self.is_output():
            raise TypeError(f"Cannot use <= to assign to output: "
                            f"{self.debug_name} (trying to assign "
                            f"{other.debug_name})")
        other = magma_value(other)
        self.wire(other)

    def __imatmul__(self, other):
        other = magma_value(other)
        if self.is_output():
            raise TypeError(f"Cannot use @= to assign to output: {self} "
                            f"(trying to assign {other})")
        self.wire(other)
        return self

    @abc.abstractmethod
    def unused(self):
        # Mark value is unused by calling unused on the underlying magma
        # elements. For example, m.Bit is wired up to a coreir term primitive A
        # general m.Array and m.Tuple will recursively call `unused` on its
        # members.
        raise NotImplementedError()

    @abc.abstractmethod
    def undriven(self):
        # Mark value is undriven by calling undriven on the underlying magma
        # elements. For example, m.Bit is wired up to a coreir undriven
        # primitive A general m.Array and m.Tuple will recursively call
        # `undriven` on its members.
        raise NotImplementedError()

    def is_driven_anon_temporary(self):
        """
        Returns true if this is an anonymous temporary value (not an output)
        that is driven
        """
        return self.name.anon() and not self.is_output() and self.driven()


class Kind(type):
    # Subclasses only need to implement one of these methods.
    def __eq__(cls, rhs):
        return cls is rhs

    __hash__ = type.__hash__

    def __repr__(cls):
        return cls.__name__

    def __str__(cls):
        return cls.__name__

    @abc.abstractmethod
    def qualify(cls, direction):
        raise NotImplementedError()

    def flip(cls):
        if cls.direction == Direction.In:
            return cls[Direction.Out]
        if cls.direction == Direction.Out:
            return cls[Direction.In]
        # Flip of inout is inout, and flip of undirected is undirected.
        return cls

    @property
    def undirected_t(cls):
        return cls.qualify(Direction.Undirected)

    @property
    def is_directed(cls):
        return cls is not cls.qualify(Direction.Undirected)


def In(T):
    return T.qualify(Direction.In)


def Out(T):
    return T.qualify(Direction.Out)


def InOut(T):
    return T.qualify(Direction.InOut)


def Flip(T):
    return T.flip()
